@@ -120,7 +120,6 @@ void
 grd_rdp_renderer_invoke_shutdown (GrdRdpRenderer *renderer)
 {
   g_assert (renderer->graphics_context);
-  g_assert (renderer->graphics_thread);
 
   stop_rendering (renderer);
 
@@ -130,6 +129,23 @@ grd_rdp_renderer_invoke_shutdown (GrdRdpRenderer *renderer)
   g_clear_pointer (&renderer->graphics_thread, g_thread_join);
 }
 
+static gpointer
+graphics_thread_func (gpointer data)
+{
+  GrdRdpRenderer *renderer = data;
+
+  if (renderer->hwaccel_nvidia)
+    grd_hwaccel_nvidia_push_cuda_context (renderer->hwaccel_nvidia);
+
+  while (!renderer->in_shutdown)
+    g_main_context_iteration (renderer->graphics_context, TRUE);
+
+  if (renderer->hwaccel_nvidia)
+    grd_hwaccel_nvidia_pop_cuda_context (renderer->hwaccel_nvidia);
+
+  return NULL;
+}
+
 void
 grd_rdp_renderer_notify_session_started (GrdRdpRenderer         *renderer,
                                          GrdRdpGraphicsPipeline *graphics_pipeline,
@@ -137,6 +153,10 @@ grd_rdp_renderer_notify_session_started (GrdRdpRenderer         *renderer,
 {
   renderer->graphics_pipeline = graphics_pipeline;
   renderer->rdp_context = rdp_context;
+
+  renderer->graphics_thread = g_thread_new ("RDP graphics thread",
+                                            graphics_thread_func,
+                                            renderer);
 }
 
 void
@@ -146,32 +166,14 @@ grd_rdp_renderer_notify_new_desktop_layout (GrdRdpRenderer *renderer,
 {
   rdpContext *rdp_context = renderer->rdp_context;
   rdpSettings *rdp_settings = rdp_context->settings;
-  uint32_t current_desktop_width =
-    freerdp_settings_get_uint32 (rdp_settings, FreeRDP_DesktopWidth);
-  uint32_t current_desktop_height =
-    freerdp_settings_get_uint32 (rdp_settings, FreeRDP_DesktopHeight);
 
-  if (renderer->graphics_pipeline)
-    renderer->pending_gfx_graphics_reset = TRUE;
-
-  if (!renderer->graphics_pipeline)
-    {
-      RdpPeerContext *rdp_peer_context = (RdpPeerContext *) rdp_context;
-
-      rfx_context_reset (rdp_peer_context->rfx_context,
-                         desktop_width, desktop_height);
-    }
-
-  if (current_desktop_width == desktop_width &&
-      current_desktop_height == desktop_height)
-    return;
+  g_assert (renderer->graphics_pipeline);
+  renderer->pending_gfx_graphics_reset = TRUE;
 
   freerdp_settings_set_uint32 (rdp_settings, FreeRDP_DesktopWidth,
                                desktop_width);
   freerdp_settings_set_uint32 (rdp_settings, FreeRDP_DesktopHeight,
                                desktop_height);
-  if (!renderer->graphics_pipeline)
-    rdp_context->update->DesktopResize (rdp_context);
 }
 
 void
@@ -425,21 +427,15 @@ grd_rdp_renderer_clear_render_contexts (GrdRdpRenderer *renderer)
   g_hash_table_remove_all (renderer->render_context_table);
 }
 
-static gpointer
-graphics_thread_func (gpointer data)
+gboolean
+grd_rdp_renderer_render_frame (GrdRdpRenderer      *renderer,
+                               GrdRdpSurface       *rdp_surface,
+                               GrdRdpRenderContext *render_context,
+                               GrdRdpBuffer        *rdp_buffer)
 {
-  GrdRdpRenderer *renderer = data;
-
-  if (renderer->hwaccel_nvidia)
-    grd_hwaccel_nvidia_push_cuda_context (renderer->hwaccel_nvidia);
-
-  while (!renderer->in_shutdown)
-    g_main_context_iteration (renderer->graphics_context, TRUE);
-
-  if (renderer->hwaccel_nvidia)
-    grd_hwaccel_nvidia_pop_cuda_context (renderer->hwaccel_nvidia);
-
-  return NULL;
+  return grd_rdp_graphics_pipeline_refresh_gfx (renderer->graphics_pipeline,
+                                                rdp_surface, render_context,
+                                                rdp_buffer);
 }
 
 GrdRdpRenderer *
@@ -452,10 +448,6 @@ grd_rdp_renderer_new (GrdSessionRdp    *session_rdp,
   renderer->session_rdp = session_rdp;
   renderer->hwaccel_nvidia = hwaccel_nvidia;
 
-  renderer->graphics_thread = g_thread_new ("RDP graphics thread",
-                                            graphics_thread_func,
-                                            renderer);
-
   return renderer;
 }
 
@@ -505,8 +497,7 @@ grd_rdp_renderer_dispose (GObject *object)
 {
   GrdRdpRenderer *renderer = GRD_RDP_RENDERER (object);
 
-  if (renderer->graphics_thread)
-    grd_rdp_renderer_invoke_shutdown (renderer);
+  grd_rdp_renderer_invoke_shutdown (renderer);
 
   if (renderer->acquired_render_contexts)
     g_assert (g_hash_table_size (renderer->acquired_render_contexts) == 0);
